from math import ceil
import tensorflow as tf

re = tf.math.real
im = tf.math.imag


@tf.function
def hot_shape(hot, length):
    return tuple(-1 if i == hot else 1 for i in range(length))


@tf.function
def alt_range(n):
    """
    Get a sequence of integers increasing in absolute value
    :param n: How many integers to get
    :return: List of alternating numbers increasing in absolute value
    """
    integers = [1] * n
    for i in range(n):
        integers[i] *= ((i + 1) // 2) if i % 2 else -((i + 1) // 2)
    return integers


@tf.function
def n_range(n):
    """
    Get a sequence of integers from -floor(n / 2) to ceil(n / 2)
    :param n: Number of integers in the resulting sequence
    :return: Tensor of integers
    """
    assert n % 2 == 1
    return tf.range(-(n // 2), ceil(n / 2))


@tf.function
def complex_conv2d(x, filters, **kwargs):
    # Shape [batch, height, width, channels, complex]

    conv = lambda v, f: tf.nn.conv2d(v, f, **kwargs)

    rex, imx = x[..., 0], x[..., 1]
    ref, imf = filters[..., 0], filters[..., 1]
    return tf.stack((conv(rex, ref) - conv(imx, imf),
                     conv(rex, imf) + conv(imx, ref)), axis=4)


@tf.function
def cs_from_weights(weights):
    # Weights have shape (1, 1, in_channels, out_channels, delta_ms, fourier_depth, 2)
    even = weights[..., 2::2, :]
    odd = weights[..., 1::2, :]
    pos_idx = tf.concat((odd, even), axis=6)
    neg_idx = tf.reverse(tf.concat((odd, -even), axis=6), axis=[5])
    idx_0 = tf.concat((weights[..., :1, :], tf.zeros_like(weights[..., :1, :])), axis=6)

    cs = tf.concat((neg_idx, idx_0, pos_idx), axis=5)

    return cs


@tf.function
def fourier(cs, ns, rs):
    """
    Fourier series from fourier coefficients cs, iteration variables ns and positions rs.
    sum_n(c_n exp(i n r))

    :param cs: Coefficients
    :param ns: Iteration variable, corresponds to that of the coefficients
    :param rs: Positions for evaluation
    :return: Tensor of values for each position in rs
    """
    nrs = ns * rs

    recs, imcs = cs[..., 0, tf.newaxis], cs[..., 1, tf.newaxis]
    real = recs * tf.cos(nrs) - imcs * tf.sin(nrs)
    imag = recs * tf.sin(nrs) + imcs * tf.cos(nrs)

    series = tf.reduce_sum(tf.concat((real, imag), axis=6), axis=5, keepdims=True)
    return series


@tf.function
def fourier_real(cs, ns, rs):
    """
    Real part of series from fourier coefficients cs, iteration variables ns and positions rs.
    sum_n(c_n exp(i n r))

    :param cs: Coefficients
    :param ns: Iteration variable, corresponds to that of the coefficients
    :param rs: Positions for evaluation
    :return: Tensor of values for each position in rs
    """
    nrs = ns * rs

    recs, imcs = cs[..., 0, tf.newaxis], cs[..., 1, tf.newaxis]
    real = recs * tf.cos(nrs) - imcs * tf.sin(nrs)
    # No imaginary part

    series = tf.reduce_sum(real, axis=5, keepdims=True)
    return series


@tf.function
def complex_mul(x, y, complex_axis: int = 6):
    rex, imx = x[..., 0], x[..., 1]
    rey, imy = y[..., 0], y[..., 1]
    return tf.stack((rex * rey - imx * imy, rex * imy + imx * rey), axis=complex_axis)


if __name__ == '__main__':
    w = tf.reshape(tf.range(20), (1, 1, 1, 1, 1, 20, 1))
    print(cs_from_weights(w))
